!---------------------------------------------------------------------------------------
! PREP_SATELLITE: average
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2022, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> average
!!
!! Purpose:    Averages individual retrievals to a grid of variable resolution
!!             where the grid resolution is adapted based on the degree of variability
!!             in the column mixing ratios
!!
!!             column average concentration over N retrievals
!!             c_av = cpri_av + sum(cak_av_i*v_av_i) + (1/N)*sum(sum(cak_ij*vpri_ij))
!!             where v_av_i is average modelled concentration each layer
!!             v_av_i = H_av_i*x*vdryair_av_i
!!
!! Interface:
!!
!!    Inputs
!!             settings  -  settings data structure
!!
!!    Outputs
!!
!!    Externals
!!
!---------------------------------------------------------------------------------------

subroutine average(settings, idate, itime, xpoints, ypoints, zpoint1, zpoint2, vmr_out, verr_out,&
                   vapri_out, cpri, cakpri, cak_out, cdryair, vdryair_out, nretr, ncoord, nlayer)

  use mod_settings
  use mod_var
  use mod_dates

  implicit none

  type (settings_t),                intent(in)        :: settings
  integer, dimension(maxretr),      intent(in out)    :: idate, itime
  real, dimension(maxretr,ncoord),  intent(in out)    :: xpoints, ypoints
  real, dimension(maxretr,nlayer),  intent(in out)    :: zpoint1, zpoint2
  real, dimension(maxretr,nlayer),  intent(in out)    :: vapri_out, vdryair_out, cak_out
  real, dimension(maxretr),         intent(in out)    :: vmr_out, verr_out
  real, dimension(maxretr),         intent(in out)    :: cdryair, cpri, cakpri
  integer,                          intent(in out)    :: nretr
  integer,                          intent(in)        :: ncoord, nlayer
  
  character(max_name_len)                             :: line
  real                                                :: llx, lly, urx, ury, xav, yav
  real, dimension(settings%nsteps)                    :: stepres
  real                                                :: res, box_mean, box_sd, box_err, rwork, area
  integer                                             :: nbox, nbox_new, num, numbox, numx, numy
  integer                                             :: ix, jy, ixy, n, nn, nr, ns, nb, nl, ierr
  integer                                             :: jjjjmmdd, hhmiss
  integer, dimension(:), allocatable                  :: retr_nbox, iretr, iretr_keep, mark_box, idate_mean, itime_mean 
  real, dimension(:), allocatable                     :: llx_box, lly_box, res_box, warea
  real, dimension(:), allocatable                     :: vmr_mean, vmr_sd, cpri_mean, cdryair_mean, cakpri_mean
  real, dimension(:,:), allocatable                   :: cak_mean, vdryair_mean, xpts_box, ypts_box, zpt1_box, zpt2_box
  integer, dimension(4)                               :: num_subbox
  real(kind=8)                                        :: jdmean

  ! if use predefined grid skip following
  ! note: predefined grid overrides domain bounds given in SETTINGS
  if ( settings%usegridfile ) then
   go to 10
  endif

  ! domain bounds
  if ( settings%lnested.eq.1 ) then
    llx = settings%outlonnest
    lly = settings%outlatnest
    urx = llx + settings%numxnest*settings%dxoutnest
    ury = lly + settings%numynest*settings%dyoutnest
  else
    llx = settings%outlonleft
    lly = settings%outlatlower
    urx = llx + settings%numxgrid*settings%dxout
    ury = lly + settings%numygrid*settings%dyout
  endif
  print*, 'llx, lly, urx, ury = ',llx,lly,urx,ury

  ! resolution steps
  stepres(1) = settings%dmin
  do ns = 2, settings%nsteps
    stepres(ns) = stepres(ns-1)*2
  end do
  print*, 'stepres = ',stepres

  ! initialize at coarsest resolution
  ! ---------------------------------

  allocate( retr_nbox(nretr) )
  allocate( iretr(nretr) )
  allocate( iretr_keep(nretr) )
  res = stepres(settings%nsteps)
!  numx = (urx - llx)/res
  numx = (urx - llx)/res + 1 ! to cover whole domain
!  numy = (ury - lly)/res
  numy = (ury - lly)/res + 1 
  nbox = numx*numy
  if ( nbox.gt.maxbox ) then
    write(logid,*) 'ERROR: maxbox too small'
    stop
  endif
  print*, 'numx, numy, nbox = ',numx, numy, nbox
  allocate( llx_box(maxbox) )
  allocate( lly_box(maxbox) )
  allocate( res_box(maxbox) )
  allocate( mark_box(maxbox) )
  allocate( warea(maxbox) )
  mark_box(:) = 0
  mark_box(1:nbox) = 1
  ! allocate retrievals to corresponding box
  do nr = 1, nretr
    ix = floor((sum(xpoints(nr,:)/4.) - llx)/res) + 1
    jy = floor((sum(ypoints(nr,:)/4.) - lly)/res) + 1
    ixy = (jy - 1)*numx + ix
    retr_nbox(nr) = ixy
    if ( nr.lt.50 ) then
      print*, 'xpoint, ypoint = ',sum(xpoints(nr,:))/4., sum(ypoints(nr,:))/4.
      print*, 'ix, jy, ixy = ',ix,jy,ixy
    endif
  end do
  ! calculate box coordinates
  do nb = 1, nbox
    jy = (nb - 1)/numx + 1
    ix = nb - (jy - 1)*numx
    llx_box(nb) = llx + (ix - 1)*res
    lly_box(nb) = lly + (jy - 1)*res
  end do
  !** test
  print*, 'box coords:'
  do nb = 1, 20
    print*, llx_box(nb), lly_box(nb)
  end do
  !**
  res_box(:) = res
  print*, 'retr_nbox(1:20) = ',retr_nbox(1:20)

  ! loop over resolution steps
  ! --------------------------

  do ns = settings%nsteps-1, 1, -1
    print*, 'STEP = ',ns
    print*, '-----------'
    res = stepres(ns)
    print*, 'res = ',res
    !--- loop over marked boxes
    do nb = 1, nbox
      if ( mark_box(nb).eq.0 ) cycle
      print*, 'nb = ',nb
      ! calculate sd of retrievals in each box
      iretr(:) = 0
      num = 0
      do nr = 1, nretr
        if ( retr_nbox(nr).eq.nb ) then
          num = num + 1
          iretr(num) = nr
        endif
      end do
      print*, 'num of retrievals in current box = ',num
      if ( num.gt.1 ) then
        box_mean = sum(vmr_out(iretr(1:num)))/real(num)
        box_sd = sqrt(sum((vmr_out(iretr(1:num))-box_mean)**2)/real(num - 1))
      else if ( num.eq.1 ) then
        box_mean = vmr_out(iretr(num))
        box_sd = 1.
      else
        box_mean = -999.
        box_sd = -999.
      endif
      print*, 'box mean, sd = ',box_mean,box_sd
      ! check if this box should be divided further
      if ( (box_sd.lt.settings%cutoff) ) then
        ! only keep if retrievals spread over all sub-boxes otherwise divide
        num_subbox(:) = 0
        do nn = 1, num
            nr = iretr(nn)
            do n = 1, 4
              jy = (n - 1)/2 + 1
              ix = n - (jy - 1)*2
              if ( sum(ypoints(nr,:)/4.).gt.(lly_box(nb)+(jy-1)*res).and.sum(ypoints(nr,:)/4.).lt.(lly_box(nb)+jy*res).and.&        
                sum(xpoints(nr,:)/4.).gt.(llx_box(nb)+(ix-1)*res).and.sum(xpoints(nr,:)/4.).lt.(llx_box(nb)+ix*res) ) then
                num_subbox(n) = num_subbox(n) + 1
              endif
            end do ! n
!          endif ! nb
        end do ! nn
        print*, 'num_subbox = ',num_subbox(:)
        if ( all(num_subbox.gt.0) ) then 
          mark_box(nb) = 0 ! do not divide box
        else if ( box_sd.eq.-999. ) then
          mark_box(nb) = 0 ! no retrievals in box
        endif
      endif
    end do ! nbox
    !--- divide boxes 
    print*, 'DIVIDE'
    print*, '------'
    nbox_new = nbox
    ! loop over boxes and divide
    do nb = 1, nbox
      if ( mark_box(nb).eq.0 ) cycle ! do not divide box
      print*, 'nb = ',nb
      ! divide box into 4 so add 3 new ones
      ! first box keeps old nb, llx and lly bounds, but update res
      res_box(nb) = res
      ! new boxes get numbers nbox_nb + n
      do n = 1, 3
        ! coordinates of new boxes
        jy = n/2
        ix = n - jy*2
        llx_box(nbox_new+n) = llx_box(nb) + ix*res
        lly_box(nbox_new+n) = lly_box(nb) + jy*res
        res_box(nbox_new+n) = res
        mark_box(nbox_new+n) = 1
        print*, 'nbox_new + n = ',nbox_new + n
        print*, 'llx_box(nb) = ',llx_box(nb)
        print*, 'llx_box(nbox_new+n) = ',llx_box(nbox_new+n)
        print*, 'lly_box(nb) = ',lly_box(nb)
        print*, 'lly_box(nbox_new+n) = ',lly_box(nbox_new+n)
      end do
      ! update retrieval box numbers
      do nr = 1, nretr
        if ( retr_nbox(nr).ne.nb ) cycle
        ix = floor((sum(xpoints(nr,:)/4.) - llx_box(nb))/res) ! not +1 because resuse nbox number for first grid in divided box
        jy = floor((sum(ypoints(nr,:)/4.) - lly_box(nb))/res)
        print*, 'update retrieval box no.s: ix, jy = ',ix,jy
        print*, 'retr_nbox(nr): old = ',retr_nbox(nr)
        if ( .not.((ix.eq.0).and.(jy.eq.0)) ) retr_nbox(nr) = nbox_new + jy*2 + ix
        print*, 'retr_nbox(nr): new = ',retr_nbox(nr)
      end do
      nbox_new = nbox_new + 3
      if ( nbox_new.gt.maxbox ) then
        write(logid,*) 'ERROR: maxbox too small'
        stop
      endif
    end do ! nb
    ! update nbox
    nbox = nbox_new
    print*, 'nbox_new = ',nbox_new
  end do ! ns

      ! TESTING
!      open(100,file=trim(settings%filegrid),action='write',status='replace',iostat=ierr)
!      write(100,*) 'llx_box lly_box res_box'
!      print*, 'llx_box(1) = ',llx_box(1)
!      do nb = 1, nbox
!        write(100,fmt='(F7.2,1X,F7.2,1X,F7.2)') llx_box(nb), lly_box(nb), res_box(nb)
!      end do
!      close(100)
!      open(100,file=trim(settings%path_obsout)//'retr_nbox_orig.txt',action='write',status='replace',iostat=ierr)
!      do nr = 1, nretr
!        write(100,fmt='(I8)') retr_nbox(nr)
!      end do
!      close(100)

10 continue
  if ( settings%usegridfile ) then
    allocate( llx_box(maxbox) )
    allocate( lly_box(maxbox) )
    allocate( res_box(maxbox) )
    allocate( retr_nbox(nretr) )
    allocate( iretr(nretr) )
    ! read llx_box, lly_box, res_box from file
    write(logid,*) 'Reading averaging grid from file'
    open(100,file=trim(settings%filegrid),status='old',action='read',iostat=ierr)
    if ( ierr.eq.0 ) then
      read(100,*) line
      do n = 1, maxbox
        read(100,fmt='(F7.2,1X,F7.2,1X,F7.2)',iostat=ierr) llx_box(n), lly_box(n), res_box(n)
        if ( ierr.ne.0 ) exit
      end do
      nbox = n - 1
      print*, 'nbox = ',nbox
      close(100)
    else
      write(logid,*) 'ERROR: file not found ',settings%filegrid
      stop
    endif
    ! allocate retrievals to corresponding box
    print*, 'nretr = ',nretr
    do nb = 1, nbox
      do nr = 1, nretr
        xav = sum(xpoints(nr,:))/4.
        yav = sum(ypoints(nr,:))/4.
        if ( (xav.gt.llx_box(nb)).and.(xav.lt.(llx_box(nb)+res_box(nb))).and. &
             (yav.gt.lly_box(nb)).and.(yav.lt.(lly_box(nb)+res_box(nb))) ) then
          retr_nbox(nr) = nb
        endif
      end do
    end do
    ! TESTING
!    open(100,file=trim(settings%path_obsout)//'retr_nbox_file.txt',action='write',status='replace',iostat=ierr)
!    do nr = 1, nretr
!      write(100,fmt='(I8)') retr_nbox(nr)
!    end do
!    close(100)
  endif ! usegridfile

  ! prepare output
  ! --------------

  allocate( vmr_mean(nbox) )
  allocate( vmr_sd(nbox) )
  allocate( cpri_mean(nbox) )
  allocate( cdryair_mean(nbox) )
  allocate( cakpri_mean(nbox) )
  allocate( cak_mean(nbox,nlayer) )
  allocate( vdryair_mean(nbox,nlayer) )
  allocate( xpts_box(nbox,ncoord) )
  allocate( ypts_box(nbox,ncoord) )
  allocate( zpt1_box(nbox,nlayer) )
  allocate( zpt2_box(nbox,nlayer) )
  allocate( idate_mean(nbox) )
  allocate( itime_mean(nbox) )

  ! loop over boxes
  numbox = 0
  do nb = 1, nbox
    ! mean and sd of retrievals in each box
    num = 0
    iretr(:) = 0
    do nr = 1, nretr
      if ( retr_nbox(nr).eq.nb ) then
        num = num + 1
        iretr(num) = nr
      endif
    end do
    if ( num.ge.settings%nmin ) then
      ! only use boxes with > nmin retrievals
      ! area weighting
      warea(:) = 0.
      do n = 1, num
        call geodarea(xpoints(iretr(n),:), ypoints(iretr(n),:), warea(n))
      end do
      warea(1:num) = warea(1:num)/sum(warea(1:num))
      print*, 'sum(warea) = ',sum(warea(1:num))
      print*, 'range(warea) = ',minval(warea(1:num)),maxval(warea(1:num))
      box_mean = dot_product(warea(1:num), vmr_out(iretr(1:num)))
      if ( num.eq.1 ) then
        box_sd = verr_out(iretr(num))
      else
        box_sd  = sqrt(sum((vmr_out(iretr(1:num))-box_mean)**2)/real(num - 1))
      endif
      box_err = sqrt(dot_product(warea(1:num), verr_out(iretr(1:num))**2))
      ! only keep retrievals within +/- 2 sd of mean
      nn = 0
      iretr_keep(:) = 0
      do n = 1, num
        if ( vmr_out(iretr(n)).gt.(box_mean-2.*box_sd).and.vmr_out(iretr(n)).lt.(box_mean+2.*box_sd) ) then
          nn = nn + 1
          iretr_keep(nn) = iretr(n)
        endif
      end do
      if (nn.eq.0) then
        ! reject this box -> no retrievals within +/- 2SD
        cycle
      endif
      ! calculate mean of kept retrievals
      warea(:) = 0.
      do n = 1, nn
        call geodarea(xpoints(iretr_keep(n),:), ypoints(iretr_keep(n),:), warea(n))
      end do
      warea(1:nn) = warea(1:nn)/sum(warea(1:nn))
      print*, 'sum(warea) = ',sum(warea(1:nn))
      print*, 'range(warea) = ',minval(warea(1:nn)),maxval(warea(1:nn))
      box_mean = dot_product(warea(1:nn), vmr_out(iretr_keep(1:nn)))
      if ( nn.eq.1 ) then
        box_sd  = vmr_out(iretr_keep(nn))      
      else
        box_sd  = sqrt(sum((vmr_out(iretr_keep(1:nn))-box_mean)**2)/real(nn - 1))      
      endif
      box_err = sqrt(dot_product(warea(1:nn), verr_out(iretr_keep(1:nn))**2))
      ! for below use num and iretr
      iretr(:) = 0
      iretr(:) = iretr_keep(:)
      ! reset and update for this box
      num = nn
      numbox = numbox + 1
    else
      ! reject this box -> too few retrievals
      cycle
    endif
    ! mean column mixing ratio
    vmr_mean(numbox) = box_mean
    ! mean column uncertainty
    vmr_sd(numbox) = box_err
    ! mean column prior convolved with averaging kernel
    rwork = 0.
    do n = 1, num
      if ( settings%satellite.eq.'tropomi' ) then
        rwork = rwork + warea(n)*dot_product(cak_out(iretr(n),:),vapri_out(iretr(n),:))
      else if (settings%satellite.eq.'bremen' ) then
        do nl = 1, nlayer
          rwork = rwork + warea(n)*vdryair_out(iretr(n),nl)*vapri_out(iretr(n),nl)*cak_out(iretr(n),nl)
        end do
      endif
    end do
    cakpri_mean(numbox) = rwork
    ! mean column prior
    cpri_mean(numbox) = dot_product(warea(1:num),cpri(iretr(1:num)))
    ! mean averaging kernel
    cak_mean(numbox,:) = matmul(warea(1:num),cak_out(iretr(1:num),:))
    ! mean column dry air concentration
    cdryair_mean(numbox) = dot_product(warea(1:num),cdryair(iretr(1:num)))
    ! mean dry air concentration
    vdryair_mean(numbox,:) = matmul(warea(1:num),vdryair_out(iretr(1:num),:))
    ! box coordinates
    xpts_box(numbox,(/1,4/)) = llx_box(nb)
    xpts_box(numbox,(/2,3/)) = llx_box(nb) + res_box(nb)
    ypts_box(numbox,(/1,2/)) = lly_box(nb)
    ypts_box(numbox,(/3,4/)) = lly_box(nb) + res_box(nb)
    zpt1_box(numbox,:) = matmul(warea(1:num),zpoint1(iretr(1:num),:))
    zpt2_box(numbox,:) = matmul(warea(1:num),zpoint2(iretr(1:num),:))
    ! date and time
    jdmean = 0d0
    do n = 1, num
      jdmean = jdmean + juldate(idate(iretr(n)),itime(iretr(n)))
    end do
    jdmean = jdmean/dble(num)
    call caldate(jdmean, jjjjmmdd, hhmiss)
    idate_mean(numbox) = jjjjmmdd
    itime_mean(numbox) = hhmiss
  end do
  print*, 'zpt1_box(1:10,:) = ',zpt1_box(1:10,:)
  nretr = numbox
  write(logid,*) 'Number of averaged retrievals = ',nretr
  print*, 'nretr = ',nretr
  vmr_out(1:nretr) = vmr_mean(1:nretr)
  verr_out(1:nretr) = vmr_sd(1:nretr)
  cdryair(1:nretr) = cdryair_mean(1:nretr)
  vdryair_out(1:nretr,:) = vdryair_mean(1:nretr,:)
  cak_out(1:nretr,:) = cak_mean(1:nretr,:)
  cpri(1:nretr) = cpri_mean(1:nretr)
  cakpri(1:nretr) = cakpri_mean(1:nretr)
  xpoints(1:nretr,:) = xpts_box(1:nretr,:)
  ypoints(1:nretr,:) = ypts_box(1:nretr,:)
  zpoint1(1:nretr,:) = zpt1_box(1:nretr,:)
  zpoint2(1:nretr,:) = zpt2_box(1:nretr,:)
  idate(1:nretr) = idate_mean(1:nretr)
  itime(1:nretr) = itime_mean(1:nretr)
  deallocate( vmr_mean )
  deallocate( vmr_sd )
  deallocate( cdryair_mean )
  deallocate( vdryair_mean )
  deallocate( cak_mean )
  deallocate( cpri_mean )
  deallocate( cakpri_mean )
  deallocate( xpts_box )
  deallocate( ypts_box )
  deallocate( zpt1_box )
  deallocate( zpt2_box )
  deallocate( warea )
  deallocate( idate_mean )
  deallocate( itime_mean )

end subroutine average

